
"""
Proton VPN Linux CLI entry point.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from importlib.metadata import version, PackageNotFoundError
import sys
from typing import List, Optional

import click
from dbus_fast.aio import MessageBus
from dbus_fast import BusType, Message, MessageType

from proton.session.exceptions import ProtonAPIError, ProtonAPINotReachable
from proton.vpn.cli._cli_constants import \
    PROGRAM_NAME, \
    HELP_OPTION, \
    HELP_OPTION_ABBREVIATED
from proton.vpn.cli.commands.account import signin, signout, info
from proton.vpn.cli.commands.server import connect, disconnect, status, servers
from proton.vpn.cli.commands.location_discovery import countries, cities
from proton.vpn.cli.commands.settings import config, SETTINGS_LIST_COMMAND
from proton.vpn.cli.core.controller import Controller, Params
from proton.vpn.cli.core.run_async import run_async
from proton.vpn.cli.core.click_exception_handler import ClickExceptionHandler


try:
    __version__ = version("proton-vpn-cli")
except PackageNotFoundError:
    __version__ = "development"

PROTON_VPN_LOGO = r"""
  ____            _               __     ______  _   _
 |  _ \ _ __ ___ | |_ ___  _ __   \ \   / /  _ \| \ | |
 | |_) | '__/ _ \| __/ _ \| '_ \   \ \ / /| |_) |  \| |
 |  __/| | | (_) | || (_) | | | |   \ V / |  __/| |\  |
 |_|   |_|  \___/ \__\___/|_| |_|    \_/  |_|   |_| \_|"""

GTK_APP_ID = "proton.vpn.app.gtk"


_CLICK_CONTEXT_SETTINGS = {"help_option_names": [HELP_OPTION, HELP_OPTION_ABBREVIATED]}


async def _vpn_gui_running() -> bool:
    bus = await MessageBus(bus_type=BusType.SESSION).connect()

    reply = await bus.call(
        Message(
            destination="org.freedesktop.DBus",
            path="/org/freedesktop/DBus",
            interface="org.freedesktop.DBus",
            member="ListNames",
        )
    )

    if reply.message_type == MessageType.ERROR:
        return False

    session_bus_names = reply.body[0]
    return GTK_APP_ID in session_bus_names


def _is_help_requested() -> bool:
    for help_flag in _CLICK_CONTEXT_SETTINGS["help_option_names"]:
        if help_flag in sys.argv:
            return True

    return False


class _OrderedGroup(click.Group):
    """OrderedGroup lists commands in the order that they were added"""
    def list_commands(self, ctx):
        return self.commands


@click.group(
    cls=_OrderedGroup,
    context_settings=_CLICK_CONTEXT_SETTINGS,
    help=f"\b {PROTON_VPN_LOGO} {__version__} \n\n Proton VPN command-line interface for Linux.",
    epilog=f"""\b
              Examples:
                {PROGRAM_NAME} {signin.name}                      # Sign in
                {PROGRAM_NAME} {connect.name}                     # Connect to fastest server
                {PROGRAM_NAME} {connect.name} --country US        # Connect to US
                {PROGRAM_NAME} {config.name} {SETTINGS_LIST_COMMAND: <20} # View settings
                {PROGRAM_NAME} {disconnect.name}                  # Disconnect
              \b
              Quick Start:
                1. Sign in:    {PROGRAM_NAME} {signin.name}
                2. Connect:    {PROGRAM_NAME} {connect.name}
                3. Disconnect: {PROGRAM_NAME} {disconnect.name}
              \b
              Get Command Help:
                {PROGRAM_NAME} [command] --help
              \b
              Documentation:
                https://protonvpn.com/support/linux-cli
              \b
              Support:
                https://protonvpn.com/support-form
                support@protonvpn.com"""
)
@click.option(
    '-v',
    '--verbose',
    help="Show detailed output during command execution",
    is_flag=True,
    default=False)
@click.pass_context
@run_async
async def app(ctx, verbose):
    """Groups all CLI commands"""
    ctx.obj.verbose = verbose
    if not ctx.obj.allow_gui_concurrency:
        command_help_requested = _is_help_requested()
        vpn_gui_running = await _vpn_gui_running()
        if not command_help_requested and vpn_gui_running:
            click.echo("Error: Proton VPN desktop app is currently running\n"
                       "The CLI and GUI cannot run simultaneously. "
                       "Please close the GUI application and try again.")
            ctx.exit()


# account related functionality
app.add_command(signin)
app.add_command(signout)
app.add_command(info)

# server related functionality
app.add_command(connect)
app.add_command(disconnect)
app.add_command(status)
app.add_command(servers)

# listing functionality
app.add_command(countries)
app.add_command(cities)

# set features
app.add_command(config)


def main(
    allow_concurrency: bool = False,
    cli_args: Optional[List[str]] = None,
    controller: Optional[Controller] = None
):
    """Runs the CLI."""
    try:
        # pylint: disable=E1120
        app(
            obj=Params(
                allow_gui_concurrency=allow_concurrency,
                overriding_controller=controller
            ),
            standalone_mode=False,
            args=cli_args
        )
    except ProtonAPIError as exc:
        click.echo(f"Error: {exc.message}", err=True)
        sys.exit(1)
    except (TimeoutError, ProtonAPINotReachable):
        click.echo(
            "Error: Network connectivity issues detected. "
            "Please check your internet connection and try again."
        )
        sys.exit(1)
    except click.Abort:
        click.echo("Abort!", err=True)
        sys.exit(1)
    except Exception as exc:
        if isinstance(exc, click.exceptions.ClickException):
            if ClickExceptionHandler.handle_error(exc):
                sys.exit(exc.exit_code)

        click.echo(
            "An unexpected error occurred. Please try again.\n"
            "If the error persists please contact customer support: "
            "https://protonvpn.com/support/contact?"
            "subject=%5BCLI%5D&os=Linux&platform=VPN%20for%20Linux"
        )
        raise  # make sure it gets reported to sentry



"""
Exceptions raised by the CLI.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""


class VPNConnectionError(Exception):
    """
    Error establishing a VPN server connection
    """


class VPNConnection2FARequiredError(VPNConnectionError):
    """
    Raised when VPN 2FA is required
    """


class CountryCodeError(Exception):
    """
    Error identifying a country by code
    """


class CountryNameError(Exception):
    """
    Error identifying a country by name
    """


class AuthenticationRequiredError(Exception):
    """
    Error performing operation that requires user authentication
    """


class AuthenticationFailedError(Exception):
    """
    Error occured during authentication
    """


class Authentication2FAFailedError(Exception):
    """
    Error occured during 2FA authentication
    """


class SignoutRequiredError(Exception):
    """
    Error performing operation requiring the user to be signed out
    """


class RequiresHigherTierError(Exception):
    """
    Requested feature that requires a higher tier
    """


class InvalidDNS(Exception):
    """Raised when user provides invalid DNS value
    """
    def __init__(self, dns: str, message: str):
        self.dns = dns
        self.message = message



"""
Helper for creating a pending async task barrier.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

import asyncio


async def wait_for_current_tasks(timeout=10):
    """
    Pauses the current task until all other tasks are processed. This is
    usefull for when we want to shutdown cleanly, or when waiting for some
    processing to complete.
    """
    try:
        future = asyncio.gather(
            *[t for t in asyncio.all_tasks()
              if t is not asyncio.current_task()]
        )
        await asyncio.wait_for(future, timeout)
    except asyncio.exceptions.TimeoutError:
        print(f"Timed out after {timeout}s waiting for tasks to finish")



"""
Exception handling module.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
import sys
from types import TracebackType
from typing import List, Optional, Protocol, Type, Union


class ExceptionReporter(Protocol):  # pylint: disable=too-few-public-methods
    """
    Protocol defining interface for reporting unhandled exceptions
    """
    def report_error(
        self,
        error: Union[
            BaseException,
            tuple[
                Optional[Type[BaseException]],
                Optional[BaseException],
                Optional[TracebackType]
            ]
        ]
    ):
        """
        Reports the provided error
        """
        raise NotImplementedError


class ExceptionHandler:
    """
    Helper class used to silently absorb and/or report specified types of unhandled exceptions
    """

    __absorbed_exceptions: List[BaseException]
    __exception_reporter: Optional[ExceptionReporter] = None

    @staticmethod
    def custom_exception_hook(exc_type, exc_value, exc_traceback):
        """
        When the exception handler is enabled, this method overrides default
        exception handling, dealing with errors that were not explicitly handled.
        :param exc_type: Type of the exception.
        :param exc_value: Instance of the exception. It might be None if the
        exception was triggered using an Exception class, rather than an object
        (e.g. raise Exception, instead of raise Exception()).
        :param exc_traceback: The exception traceback.
        """
        absorb = False
        for absorb_exc in ExceptionHandler.__absorbed_exceptions:
            if issubclass(exc_type, absorb_exc):
                absorb = True
                break

        if not absorb:
            if ExceptionHandler.__exception_reporter and \
               ExceptionHandler.__is_reportable_exception(exc_type):
                ExceptionHandler.__exception_reporter.report_error(
                    (exc_type, exc_value, exc_traceback)
                )

            sys.exit(1)

    @staticmethod
    def __is_reportable_exception(exc_type) -> bool:
        if issubclass(exc_type, AssertionError):
            # We shouldn't catch assertion errors raised by tests.
            return False

        if not issubclass(exc_type, Exception):
            return False

        return True

    @staticmethod
    def asyncio_exception_handler(_, context):
        """
        When the exception handler is enabled, this method overrides default
        asyncio current event loop exception handling,
        dealing with errors that were not explicitly handled.
        """
        exception = context.get('exception', None)
        if exception:
            exc_info = (type(exception), exception, exception.__traceback__)
            ExceptionHandler.custom_exception_hook(*exc_info)

    @staticmethod
    def set_uncaught_exceptions_to_absorb(exceptions: List[BaseException]):
        """
        Silences list of provided exceptions if raised and uncaught
        """
        ExceptionHandler.__absorbed_exceptions = exceptions

    @staticmethod
    def enable(exception_reporter: Optional[ExceptionReporter] = None):
        """
        Enable global handling/absorption/reporting of uncaught exceptions
        """
        ExceptionHandler.__exception_reporter = exception_reporter
        ExceptionHandler.__absorbed_exceptions = []

        sys.excepthook = ExceptionHandler.custom_exception_hook
        asyncio.get_event_loop().set_exception_handler(
            ExceptionHandler.asyncio_exception_handler
        )

    @staticmethod
    def disable():
        """
        Disable global handling/absorption/reporting of uncaught exceptions
        """
        sys.excepthook = sys.__excepthook__
        asyncio.get_event_loop().set_exception_handler(None)  # resets to the default handler



"""
Feature command definitions for config set subcommands.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Protocol

from proton.vpn.killswitch.interface import KillSwitchState
from proton.vpn.core.settings.features import NetShield
from proton.vpn.core.settings.custom_dns import CustomDNS
from proton.vpn.cli.core.controller import Feature


class ClickArgType(Protocol):
    """Maps between VPN state types and click friendly formats"""

    @staticmethod
    @abstractmethod
    def get_human_friendly_state_string(value: Any) -> str:
        """Returns human friendly state string for the specified value."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_list_of_str() -> list[str]:
        """Converts to a list of all possible click value strings

        This is a necessary change because <8.2.0, `click.Choice`
        can only take a list of strings. With version >=8.2.0 click
        supports non-string choices (you can pass an enum class).
        This change is to make it backwards compatible, as on Fedora 43
        click is v8.1.7 and on Ubuntu 24.04 v8.1.6.
        See more here: https://click.palletsprojects.com/en/stable/api/#click.Choice
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def from_str(value: str) -> Any:
        """Returns value based on provided click string"""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_str(value: Any) -> str:
        """Returns click string based on provided value"""
        raise NotImplementedError


class ToggleType(ClickArgType):
    """Represents simple binary options that a user can select."""

    @staticmethod
    def get_human_friendly_state_string(value: bool) -> str:
        """Returns human friendly state string for the specified value."""
        if value is False:
            return "disabled"

        return "enabled"

    @staticmethod
    def to_list_of_str() -> list[str]:
        """Converts to a list of all possible click value strings"""
        return ["off", "on"]

    @staticmethod
    def from_str(value: str) -> bool:
        """Returns value based on provided click string"""
        if value.lower() == "off":
            return False

        return True

    @staticmethod
    def to_str(value: bool) -> str:
        """Returns click string based on provided value"""
        if value is False:
            return "off"

        return "on"


class CustomDNSType(ToggleType):
    """Toggle extension for CustomDNS"""

    @staticmethod
    def to_str(value: CustomDNS) -> str:
        """Returns click string based on provided value"""
        enabled = ToggleType.to_str(value.enabled)

        max_ip_list_length = 2
        if value.ip_list:
            ips = value.ip_list[:max_ip_list_length]
            ips_to_show = [str(dns.ip) for dns in ips]
            if len(value.ip_list) > max_ip_list_length:
                ips_to_show.append("...")
            return f"{enabled}  [{', '.join(ips_to_show)}]"

        # no DNS IPs to show
        return enabled


class KillSwitchType(ClickArgType):
    """Represents the various kill switch states that a user can select"""
    # Advanced option is temporarily removed as
    # currently it's not possible to establish a connection with it enabled,
    # while being disconnected.

    @staticmethod
    def get_human_friendly_state_string(value: KillSwitchState) -> str:
        """Returns human friendly state string for the specified value."""
        if value == KillSwitchState.OFF:
            return "disabled"

        return "standard"

    @staticmethod
    def to_list_of_str() -> list[str]:
        """Converts to a list of all possible click value strings

        See explanation above.
        """
        return ["off", "standard"]

    @staticmethod
    def from_str(value: str) -> KillSwitchState:
        """Returns value based on provided click string"""
        if value.upper() == KillSwitchState.OFF.name:
            return KillSwitchState.OFF

        return KillSwitchState.ON

    @staticmethod
    def to_str(value: KillSwitchState) -> str:
        """Returns click string based on provided value"""
        if value == KillSwitchState.OFF:
            return "off"

        return "standard"


class NetshieldType(ClickArgType):
    """Represent the various netshield states that a user can select"""
    @staticmethod
    def get_human_friendly_state_string(value: NetShield) -> str:
        """Returns human friendly state string for the specified value."""
        if value == NetShield.NO_BLOCK:
            return "disabled"

        if value == NetShield.BLOCK_MALICIOUS_URL:
            return "malware only"

        return "malware, ads and trackers"

    @staticmethod
    def to_list_of_str() -> list[str]:
        """Converts to a list of all possible click value strings"""
        return ["off", "malware-only", "malware-ads-trackers"]

    @staticmethod
    def from_str(value: str) -> NetShield:
        """Returns value based on provided click string"""
        if value == "off":
            return NetShield.NO_BLOCK

        if value == "malware-only":
            return NetShield.BLOCK_MALICIOUS_URL

        return NetShield.BLOCK_ADS_AND_TRACKING

    @staticmethod
    def to_str(value: NetShield) -> str:
        """Returns click string based on provided value"""
        if value == NetShield.NO_BLOCK:
            return "off"

        if value == NetShield.BLOCK_MALICIOUS_URL:
            return "malware-only"

        return "malware-ads-trackers"


@dataclass
class ClickFeature(Feature):
    """Click specific feature data."""
    command: str = None
    human_friendly_name: str = None
    available_setting_description: str = None
    help_description: str = None
    help_epilog: str = None
    value_to_help: dict[str, str] = None
    click_type: ClickArgType = None


VPN_ACCELERATOR_FEATURE = ClickFeature(
    command="vpn-accelerator",
    human_friendly_name="VPN Accelerator",
    setting_path="features.vpn_accelerator",
    available_setting_description="Performance optimization",
    help_description="Enable or disable VPN Accelerator for improved connection speeds.",
    value_to_help={
        "off": "Disable VPN Accelerator",
        "on": " Enable VPN Accelerator (recommended)"
    },
    help_epilog="""\b
Values:
{feature_values}
\b
Examples:
  {program_name} {config_command} {set_command} {feature_command} on
  {program_name} {config_command} {set_command} {feature_command} off
\b
Current Setting:
  Run '{program_name} {config_command} {settings_list_command}' to see current value.
\b
How It Works:
  VPN Accelerator uses advanced technologies to increase VPN speeds
  by up to 400% without compromising security. It optimizes the
  connection between your device and Proton VPN servers.
\b
Recommendation:
  Keep enabled for best performance. Only disable if experiencing
  connection issues with specific networks.""",
    click_type=ToggleType()
)


MODERATE_NAT_FEATURE = ClickFeature(
    command="moderate-nat",
    human_friendly_name="Moderate NAT",
    setting_path="features.moderate_nat",
    available_setting_description="NAT type for gaming/P2P",
    help_description="Enable or disable Moderate NAT for improved gaming and P2P performance.",
    value_to_help={
        "off": "Disable Moderate NAT (Strict NAT)",
        "on": " Enable Moderate NAT"
    },
    help_epilog="""\b
Values:
{feature_values}
\b
Examples:
  {program_name} {config_command} {set_command} {feature_command} on
  {program_name} {config_command} {set_command} {feature_command} off
\b
Current Setting:
  Run '{program_name} {config_command} {settings_list_command}' to see current value.
\b
How It Works:
  NAT (Network Address Translation) type affects your ability to connect
  to other players in online games and peers in P2P networks.

  Strict NAT (off): Maximum security, may limit connections in games/P2P
  Moderate NAT (on): Better connectivity for gaming and P2P, still secure""",
    click_type=ToggleType()
)


IPV6_FEATURE = ClickFeature(
    command="ipv6",
    human_friendly_name="IPv6",
    setting_path="ipv6",
    available_on_free_tier=True,
    requires_restart=True,
    available_setting_description="IPv6 support",
    help_description="Enable or disable IPv6 support.",
    value_to_help={
        "off": "Disable IPv6 (default)",
        "on": " Enable IPv6"
    },
    help_epilog="""\b
Values:
{feature_values}
\b
Examples:
  {program_name} {config_command} {set_command} {feature_command} on
  {program_name} {config_command} {set_command} {feature_command} off
\b
Current Setting:
  Run '{program_name} {config_command} {settings_list_command}' to see current value.
\b
How It Works:
  When enabled, IPv6 traffic will be routed through the VPN tunnel.""",
    click_type=ToggleType()
)


ANON_CRASH_REPORTS_FEATURE = ClickFeature(
    command="anonymous-crash-reports",
    human_friendly_name="Anonymous crash reports",
    setting_path="anonymous_crash_reports",
    available_on_free_tier=True,
    available_setting_description="Anonymous crash reporting",
    help_description="Enable or disable anonymous crash reports.",
    value_to_help={
        "off": "Disable crash reporting",
        "on": " Enable anonymous crash reporting"
    },
    help_epilog="""\b
Values:
{feature_values}
\b
Examples:
  {program_name} {config_command} {set_command} {feature_command} on
  {program_name} {config_command} {set_command} {feature_command} off
\b
Current Setting:
  Run '{program_name} {config_command} {settings_list_command}' to see current value.
\b
How It Works:
  When enabled, the CLI sends anonymous crash reports to help us:
  - Fix bugs and improve stability
  - Detect firewall interference
  - Avoid VPN blocks
\b
Privacy:
  - Reports are completely anonymous
  - No personal information is collected
  - Helps improve the CLI for everyone""",
    click_type=ToggleType()
)


PORT_FORWARDING_FEATURE = ClickFeature(
    command="port-forwarding",
    human_friendly_name="Port forwarding",
    setting_path="features.port_forwarding",
    available_setting_description="Enable port forwarding for P2P",
    help_description="Enable or disable port forwarding for P2P applications.",
    value_to_help={
        "off": "Disable port forwarding",
        "on": " Enable port forwarding"
    },
    help_epilog="""\b
Values:
{feature_values}
\b
How It Works:
  When enabled, the VPN client requests a forwarded port during
  connection to P2P-capable servers. The port assignment requires
  an external script to maintain the lease and retrieve the port
  number. Without the script, the assigned port expires.
\b
Setup Guide:
  https://protonvpn.com/support/port-forwarding-manual-setup#linux
\b
Examples:
  {program_name} {config_command} {set_command} {feature_command} on
  {program_name} {config_command} {set_command} {feature_command} off""",
    click_type=ToggleType()
)


CUSTOM_DNS_FEATURE = ClickFeature(
    command="custom-dns",
    human_friendly_name="Custom DNS",
    setting_path="custom_dns",
    requires_restart=True,
    available_setting_description="Use custom DNS servers",
    help_description="Configure custom DNS servers.",
    value_to_help={
        "off": "Disable custom DNS (use Proton's DNS)",
        "on": " Enable custom DNS with specified servers"
    },
    help_epilog="""\b
Values:
{feature_values}
\b
Examples:
  # Enable with custom DNS
  {program_name} {config_command} {set_command} {feature_command} on --dns 1.1.1.1,8.8.8.8
\b
  # Enable with IPv6
  {program_name} {config_command} {set_command} {feature_command} on --dns 2606:4700:4700::1111
\b
  # Disable custom DNS
  {program_name} {config_command} {set_command} {feature_command} off
\b
Current Setting:
  Run '{program_name} {config_command} {settings_list_command}' to see current DNS configuration.
""",
    click_type=CustomDNSType(),
)


NETSHIELD_FEATURE = ClickFeature(
    command="netshield",
    human_friendly_name="NetShield",
    setting_path="features.netshield",
    available_setting_description="Ad-blocking and malware protection",
    help_description="Configure NetShield ad-blocking and malware protection.",
    value_to_help={
        "off": "                 NetShield disabled",
        "malware-only": "        Block malware domains",
        "malware-ads-trackers": "Block malware, ads, and trackers (recommended)"
    },
    help_epilog="""\b
Values:
{feature_values}
\b
Examples:
  {program_name} {config_command} {set_command} {feature_command} off
  {program_name} {config_command} {set_command} {feature_command} malware-only
  {program_name} {config_command} {set_command} {feature_command} malware-ads-trackers
\b
Current Setting:
  Run '{program_name} {config_command} {settings_list_command}' to see current value.
\b
How It Works:
  NetShield blocks malicious and advertising domains at the DNS level,
  preventing malware downloads and removing ads before they load.""",
    click_type=NetshieldType()
)


KILLSWITCH_FEATURE = ClickFeature(
    command="kill-switch",
    human_friendly_name="Kill switch",
    setting_path="killswitch",
    available_on_free_tier=True,
    available_setting_description="Block internet if VPN drops",
    help_description="Configure Kill Switch to block internet if VPN connection drops.",
    value_to_help={
        "off": "     Disable Kill Switch (internet always accessible)",
        "standard": "Block internet only while VPN is active"
    },
    help_epilog="""\b
Values:
{feature_values}
\b
Examples:
  {program_name} {config_command} {set_command} {feature_command} off
  {program_name} {config_command} {set_command} {feature_command} standard
\b
Current Setting:
  Run '{program_name} {config_command} {settings_list_command}' to see current value.
\b
Behavior:
  - standard: Blocks internet if VPN drops during active connection
              Restores internet when you disconnect intentionally""",
    click_type=KillSwitchType()
)

BOOL_FEATURES = [
    VPN_ACCELERATOR_FEATURE,
    MODERATE_NAT_FEATURE,
    IPV6_FEATURE,
    ANON_CRASH_REPORTS_FEATURE,
    PORT_FORWARDING_FEATURE,
]

ALL_FEATURES = [
    NETSHIELD_FEATURE,
    KILLSWITCH_FEATURE,
    PORT_FORWARDING_FEATURE,
    CUSTOM_DNS_FEATURE,
    VPN_ACCELERATOR_FEATURE,
    MODERATE_NAT_FEATURE,
    IPV6_FEATURE,
    ANON_CRASH_REPORTS_FEATURE,
]



"""
Program name resolution.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import sys

from proton.utils.environment import VPNExecutionEnvironment

PROGRAM_NAME = os.path.basename(sys.argv[0]) if sys.argv else "protonvpn"

LOGGING_FILENAME = "vpn-cli"
LOGGING_DIR_PATH = os.path.join(VPNExecutionEnvironment().path_cache, "logs")

HELP_OPTION = "--help"
HELP_OPTION_ABBREVIATED = "-h"



"""
Utilities to facilitate command implementation.

Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import click

from proton.vpn.cli.core.controller import Controller


async def inform_that_expired_serverlist_will_be_updated_if_necessary(controller: Controller):
    """Warns user of server list update if it has expired"""
    if await controller.is_serverlist_expired():
        click.echo("Server list is outdated, updating... This may take a moment.")



"""
Server/Connection related commands.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import click
from tabulate import tabulate

from proton.vpn.cli._cli_constants import PROGRAM_NAME
from proton.vpn.cli.core.run_async import run_async
from proton.vpn.cli.core.controller import Controller
from proton.vpn.session.servers.types import ServerFeatureEnum
from proton.vpn.cli.core.exceptions import \
    AuthenticationRequiredError, \
    CountryCodeError, \
    CountryNameError
from proton.vpn.cli.commands.account import SIGNIN_COMMAND
from proton.vpn.cli.commands.command_utils import \
    inform_that_expired_serverlist_will_be_updated_if_necessary
from proton.vpn.cli.commands.server import CONNECT_COMMAND

COUNTRIES_COMMAND = "countries"
COUNTRIES_LIST_COMMAND = "list"
CITIES_COMMAND = "cities"
CITIES_LIST_COMMAND = "list"
COUNTRY_ARGUMENT = "COUNTRY"

FEATURES_TO_DISPLAY = {
    ServerFeatureEnum.P2P: "P2P",
    ServerFeatureEnum.SECURE_CORE: "Secure Core",
    ServerFeatureEnum.TOR: "Tor",
}


def _print_usage_error(msg: str):
    raise click.UsageError(msg)


@click.group(
    name=COUNTRIES_COMMAND,
    no_args_is_help=False,
    epilog=f"""\b
Output:
  Displays country names with their two-letter codes and server counts.
\b
Examples:
  {PROGRAM_NAME} {COUNTRIES_COMMAND} {COUNTRIES_LIST_COMMAND}
\b
Connect to a country:
  {PROGRAM_NAME} {CONNECT_COMMAND} --country <CODE>
\b
Explore cities in a country:
  {PROGRAM_NAME} {CITIES_COMMAND} {CITIES_LIST_COMMAND} <COUNTRY>"""
)
@run_async
async def countries():
    """Discover all countries where Proton VPN servers are available."""


@countries.command(
    name=COUNTRIES_LIST_COMMAND,
    short_help="List all countries where Proton VPN servers are available."
)
@click.pass_context
@run_async
async def list_countries(ctx):
    """List all countries where Proton VPN servers are available."""
    controller = await Controller.create(params=ctx.obj, click_ctx=ctx)

    await inform_that_expired_serverlist_will_be_updated_if_necessary(controller)

    try:
        all_countries = await controller.get_all_countries()
    except AuthenticationRequiredError:
        _print_usage_error(
            "Authentication required to view complete country list. "
            f"Please sign in with '{controller.program_name} {SIGNIN_COMMAND}'"
        )
        return

    table = tabulate(
        [(country.name, country.code.upper()) for country in all_countries],
        headers=["Country", "Code"],
        tablefmt="simple",
        stralign="left",
        numalign="right",
    )
    click.echo(table)


@click.group(
    name=CITIES_COMMAND,
    no_args_is_help=False,
    short_help="Discover cities where servers are available in a specific country.",
    epilog=f"""
\b
Country Format:
  Two-letter code: US, GB, DE
  Full name: "United States", "Germany" (use quotes for multi-word)
\b
Examples:
  {PROGRAM_NAME} {CITIES_COMMAND} {CITIES_LIST_COMMAND} US               # Cities in US (direct argument)
  {PROGRAM_NAME} {CITIES_COMMAND} {CITIES_LIST_COMMAND} "United States"  # Full country name
\b
Connect to a city:
  {PROGRAM_NAME} {CONNECT_COMMAND} --city <CITY>
\b
To see available countries:
  {PROGRAM_NAME} {COUNTRIES_COMMAND} {COUNTRIES_LIST_COMMAND}"""  # noqa:E501
)
@run_async
async def cities():
    """Discover cities where servers are available in a specific country."""


@cities.command(
    name=CITIES_LIST_COMMAND,
    short_help="Shows a list of cities where servers are available in a specific country."
)
@click.argument(COUNTRY_ARGUMENT, required=True)
@click.pass_context
@run_async
async def list_cities_in_country(ctx, country: str):
    """
    Shows a list of cities where servers are available in a specific country.

    Arguments:

    COUNTRY    Country code (US, GB, DE) or full name ("United States")
    """
    controller = await Controller.create(params=ctx.obj, click_ctx=ctx)

    await inform_that_expired_serverlist_will_be_updated_if_necessary(controller)

    try:
        all_countries = await controller.get_all_countries()
    except AuthenticationRequiredError:
        _print_usage_error(
            "Authentication required to view cities. "
            f"Please sign in with '{controller.program_name} {SIGNIN_COMMAND}'"
        )

    try:
        country_code = controller.validate_country_input(country)
    except CountryCodeError:
        _print_usage_error(
            f"Invalid country code '{country}'. Please use a valid country code."
        )
    except CountryNameError:
        _print_usage_error(
            f"Invalid country name '{country}'. Please use a valid country name."
        )

    requested_country = \
        list(filter(lambda country: country.code == country_code.lower(), all_countries))

    if not requested_country:
        _print_usage_error(
            f"Country '{country}' not found. "
            f"Use '{controller.program_name} {COUNTRIES_COMMAND}' to see available options."
        )

    requested_country = requested_country.pop()
    table_data = []

    for city in requested_country.locations:
        only_displayable_features = [
            feature_display_name
            for feature, feature_display_name in FEATURES_TO_DISPLAY.items()
            if feature in city.features
        ]
        sorted_human_readable_features = sorted(only_displayable_features)

        table_data.append((city.name, ", ".join(sorted_human_readable_features)))

    table = tabulate(
        table_data,
        headers=["City", "Features"],
        tablefmt="simple",
        stralign="left",
        numalign="right",
    )

    click.echo(f"\nCities in {requested_country.name}:\n{table}\n")



"""
Core CLI logic.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
from contextlib import asynccontextmanager
from dataclasses import dataclass
from importlib import metadata
import random
from types import TracebackType
from typing import Callable, List, Optional, Type, Union

from click.core import Context as ClickContext
from packaging.version import Version
import sentry_sdk

from proton.session.exceptions import ProtonAPIAuthenticationNeeded
from proton.vpn.core.settings.custom_dns import CustomDNSEntry, CustomDNS
from proton.vpn import logging as ProtonLogging
from proton.vpn.cli._cli_constants import LOGGING_FILENAME, LOGGING_DIR_PATH
from proton.vpn.cli.core.exception_handler import ExceptionHandler
from proton.vpn.cli.core.exceptions import \
    Authentication2FAFailedError, \
    AuthenticationRequiredError, \
    AuthenticationFailedError, \
    CountryCodeError, \
    CountryNameError, \
    InvalidDNS, \
    RequiresHigherTierError, \
    SignoutRequiredError, \
    VPNConnection2FARequiredError, \
    VPNConnectionError
from proton.vpn.connection import states, events
from proton.vpn.connection.enum import ConnectionStateEnum
from proton.vpn.core.api import ProtonVPNAPI
from proton.vpn.core.vpnconnector import VPNStateSubscriber, VPNConnection, VPNConnector
from proton.vpn.core.session_holder import ClientTypeMetadata
from proton.vpn.core.settings import Settings
from proton.vpn.session import ServerList
from proton.vpn.session.dataclasses.servers import Country
from proton.vpn.session.servers.country_codes import \
    validate_country_code, \
    get_country_code_for_name
from proton.vpn.session.servers.types import LogicalServer, ServerFeatureEnum

DEFAULT_CLI_NAME = "protonvpn"


@dataclass
class Feature:
    """Used when setting and saving features."""
    setting_path: str = None
    available_on_free_tier: bool = False
    requires_restart: bool = False


@dataclass
class Params:
    """The parameters for constructing the Controller"""
    verbose: str = False
    allow_gui_concurrency: bool = False
    overriding_controller: Optional['Controller'] = None


@asynccontextmanager
async def _wait_for_event(  # pylint: disable=R0913
    connector: VPNConnector,
    event_types: Optional[List[ConnectionStateEnum]] = None,
    timeout: int = 10,
    wait_for_new_connection: Optional[bool] = False
):
    if not event_types:
        yield
        return

    event = asyncio.Event()
    current_connection = connector.current_connection

    class Subscriber(VPNStateSubscriber):  # pylint: disable=R0903
        """
        This class listens for a given set of status changes and then
        triggers the given event.
        """
        event_hit_count: int
        connection_state: states.State

        def status_update(self, status: states.State):
            if status.type in event_types:
                if wait_for_new_connection:
                    if current_connection is status.context.connection:
                        return

                self.connection_state = status
                event.set()
            elif status.type is ConnectionStateEnum.ERROR:
                self.connection_state = status
                event.set()

        @property
        def error_event_occurred(self):
            """Returns whether a connection error event has occurred."""
            return isinstance(self.connection_state, states.Error)

    subscriber = Subscriber()
    connector.register(subscriber)

    yield

    try:
        await asyncio.wait_for(event.wait(), timeout)

    except asyncio.exceptions.TimeoutError as exc:
        connector.unregister(subscriber)
        expected_types = ", ".join(ConnectionStateEnum(type).name for type in event_types)  # noqa: E501 # pylint: disable=C0301
        logger = ProtonLogging.getLogger(__name__)
        logger.error(f"Timed out after {timeout}s waiting for event(s): {expected_types}")
        raise TimeoutError from exc

    connector.unregister(subscriber)
    if subscriber.error_event_occurred:
        error_event = subscriber.connection_state.context.event
        if isinstance(error_event, events.TwoFARequired):
            raise VPNConnection2FARequiredError
        raise VPNConnectionError


class Controller:  # pylint: disable=too-many-public-methods
    """
    The application business logic is in this class. The is the core of the
    application.
    """
    def __init__(
        self,
        params: Params,
        click_ctx: ClickContext,
        api: ProtonVPNAPI = None
    ):
        ProtonLogging.config(
            filename=LOGGING_FILENAME,
            logdirpath=LOGGING_DIR_PATH,
            log_to_console=params.verbose
        )

        client_type_metadata = ClientTypeMetadata(
            type="cli"
        )

        ExceptionHandler.enable(exception_reporter=self)
        self._api = api or ProtonVPNAPI(client_type_metadata)
        self._click_context = click_ctx

    @staticmethod
    async def create(params: Params, click_ctx: ClickContext):
        """Preferred method to get an instance of Controller."""
        controller = params.overriding_controller or Controller(params, click_ctx)
        # Ensure controller always has settings loaded,
        # even if only using free user defaults prior to authentication.
        # This allows crash reporting to work prior to sign in.
        await controller.get_settings()
        return controller

    def set_uncaught_exceptions_to_absorb(self, exceptions: List[BaseException]):
        """
        Silences list of provided exceptions if raised and uncaught
        """
        ExceptionHandler.set_uncaught_exceptions_to_absorb(exceptions)

    # ExceptionReporter protocol
    def report_error(
        self,
        error: Union[
            BaseException,
            tuple[
                Optional[Type[BaseException]],
                Optional[BaseException],
                Optional[TracebackType]
            ]
        ]
    ):
        """Sends the error to Sentry."""
        self._api.usage_reporting.report_error(error)
        sentry_version = Version(metadata.version("sentry-sdk"))
        if sentry_version < Version("2.0.0"):
            sentry_sdk.flush()
        else:
            sentry_sdk.get_client().flush()  # pylint: disable=no-member

    @property
    def program_name(self) -> Optional[str]:
        """Returns the name of the CLI"""
        return self._click_context.find_root().info_name

    @property
    def is_logged_in(self) -> bool:
        """Returns whether the user is logged in or not"""
        return self._api.is_user_logged_in()

    @property
    def user_tier(self) -> int:
        """Returns the Proton VPN tier"""
        return self._api.user_tier

    @property
    def user_on_free_tier(self) -> bool:
        """Returns if the current user is on free tier or not.
        """
        return self.user_tier == 0

    @property
    def account_name(self) -> str:
        """
        Returns the Proton VPN account name if currently logged in.
        """
        return self._api.account_name

    async def get_settings(self) -> Settings:
        """Returns general settings."""
        return await self._api.load_settings()

    async def save_feature_setting(
        self,
        feature: Feature,
        value: Union[int, bool, CustomDNS]
    ):
        """Ensures that the feature is stored to disk only
        if the subscription tier allows it.

        Args:
            feature (Feature): feature whose setting will be modified
            value: value to set for the given feature

        Raises:
            AuthenticationRequiredError: if user is not logged in
            RequiresHigherTierError: if feature requires a higher plan
        """
        if not self.is_logged_in:
            raise AuthenticationRequiredError

        settings = await self.get_settings()

        if not feature.available_on_free_tier and self.user_on_free_tier:
            raise RequiresHigherTierError

        settings = self._set_settings_by_path(settings, feature.setting_path, value)

        await self.save_settings(settings)

    async def get_feature_setting(
        self,
        feature: Feature
    ) -> Union[int, bool, CustomDNS]:
        """Retrieves the current setting for the requested feature.

        Args:
            feature (Feature): feature whose setting will be returned

        Raises:
            AuthenticationRequiredError: if user is not logged in
        """
        if not self.is_logged_in:
            raise AuthenticationRequiredError

        settings = await self.get_settings()

        return self._get_settings_by_path(settings, feature.setting_path)

    async def save_settings(self, settings: Settings):
        """Saves general settings."""
        connector = await self.get_vpn_connector()
        # Paid users always push feature changes to a live connection; wait for
        # the LA CONNECTED event confirming completion. Free users never do.
        wait_for_feature_apply = not self.user_on_free_tier and connector.is_connected
        event_types = [ConnectionStateEnum.CONNECTED] if wait_for_feature_apply else None
        async with _wait_for_event(connector, event_types=event_types):
            await self._api.save_settings(settings)

    def _set_settings_by_path(
        self,
        settings: Settings,
        path: str,
        value: Union[int, bool, CustomDNS]
    ) -> Settings:
        parts = path.split(".")
        current = settings
        for part in parts[:-1]:
            current = getattr(current, part)

        setattr(current, parts[-1], value)
        return settings

    def _get_settings_by_path(
        self,
        settings: Settings,
        path: str
    ) -> Union[int, bool, CustomDNS]:
        parts = path.split(".")
        current = settings
        for part in parts[:-1]:
            current = getattr(current, part)

        return getattr(current, parts[-1])

    def parse_dns_ips(self, dns_list: list[str]) -> list[CustomDNSEntry]:
        """Parses a CSV string of DNS IPs into a list of strings.

        Args:
            dns_csv (str): CSV string of DNS IPs.

        Returns:
            List[str]: List of DNS IPs.

        Raises:
            ValueError: If any of the DNS IPs are invalid.
        """
        parsed_dns_list = []
        for dns in dns_list:
            try:
                parsed_dns_list.append(CustomDNSEntry.new_from_string(dns))
            except ValueError as excp:
                raise InvalidDNS(dns, excp) from excp
        return parsed_dns_list

    def to_custom_dns(self, state: bool, dns_ips: list[CustomDNSEntry]) -> CustomDNS:
        """Convert to custom dns setting object.

        Args:
            state (bool): whether is enabled or disabled
            dns_ips (list[CustomDNSEntry]): list of IPs

        Returns:
            CustomDNS: created object
        """
        return CustomDNS(state, dns_ips)

    async def get_current_connection(self) -> Optional[VPNConnection]:
        """Returns the current VPN connection or None if there isn't one."""
        return (await self.get_vpn_connector()).current_connection

    async def is_connection_active(self) -> bool:
        """
        Returns whether the current connection is active or not.

        A connection is considered active in the connecting, connected
        and disconnecting states.
        """
        return (await self.get_vpn_connector()).is_connection_active

    # pylint: disable=too-many-arguments
    async def find_logical_server(
        self,
        server_name: Optional[str] = None,
        country: Optional[str] = None,
        city: Optional[str] = None,
        features: ServerFeatureEnum = 0,
        random_server: bool = False
    ) -> Optional[LogicalServer]:
        """
        Finds a server in the serverlist meeting the user's criteria
        :param server_name: The name of the server to connect to.
        :param country: The country whose fastest/random server we want to connect to.
        :param city: The city whose fastest/random server we want to connect to.
        :param features: The required features of the fastest/random server we wish to connect to.
        :param random_server: If true, look for a random (instead of fastest) server
                       meeting all requirements.
        :return: The fastest logical server meeting the provided constraints.
        """
        if not self._api.is_user_logged_in():
            raise AuthenticationRequiredError

        free_user = self.user_tier == 0
        requesting_paying_feature =\
            (server_name or country or city or features or random_server)
        if free_user and requesting_paying_feature:
            raise RequiresHigherTierError

        logical_server = None
        server_list = await self.get_updated_server_list()

        # server name takes precedence
        if server_name:
            logical_server = server_list.get_by_name(server_name)
        else:
            servers = server_list.logicals

            # location filtering
            # check if we're looking in a city
            if city:
                servers = ServerList.get_servers_in_city(servers, city)
            # or see if we're looking in a country
            elif country:
                valid_country_code = self.validate_country_input(country)
                servers = ServerList.get_servers_in_country_code(servers, valid_country_code)

            # feature filtering
            features_excluded_by_default =\
                ServerFeatureEnum.SECURE_CORE | ServerFeatureEnum.TOR

            # don't exclude features that are explicitly requested
            features_excluded_by_default =\
                (features_excluded_by_default & features) ^ features_excluded_by_default

            servers = ServerList.get_servers_with_features(
                servers,
                request_features=features,
                exclude_features=features_excluded_by_default
            )

            # grab available servers
            servers = ServerList.get_available_servers(servers, self.user_tier)

            # random server or fastest server
            if random_server:
                filtered_servers = list(servers)
                if len(filtered_servers) > 0:
                    logical_server =\
                        random.choice(filtered_servers)  # nosec B311 # nosemgrep: gitlab.bandit.B311 # noqa: E501 # pylint: disable=line-too-long
            else:
                logical_server = ServerList.get_fastest_server(servers)

        return logical_server

    async def get_all_countries(self) -> List[Country]:
        """Returns a list of countries."""
        if not self._api.is_user_logged_in():
            raise AuthenticationRequiredError

        server_list = await self.get_updated_server_list()
        return server_list.group_by_country(group_by_city=True)

    async def connect(
        self,
        server: LogicalServer
    ) -> Optional[states.State]:
        """
        Establishes a VPN connection.
        :param server: The specified server to connect to.
        :return: Connection state on successful connection, None otherwise
        """
        if not self._api.is_user_logged_in():
            raise AuthenticationRequiredError

        # make sure our certificate hasn't expired, or isn't about to.
        # this needs to happen before we establish connection state
        # otherwise if we are in a connected/error state then we will block
        # after starting/restarting local agent listener synchronously
        await self._api.refresher.update_certificate_if_necessary()

        connector = await self.get_vpn_connector()

        try:
            async with _wait_for_event(
                connector,
                event_types=[ConnectionStateEnum.CONNECTED],
                wait_for_new_connection=True
            ):
                await self._connect(server)
        except VPNConnection2FARequiredError:
            raise
        except (TimeoutError, VPNConnectionError):
            # If the connection fails, clean up NM setup
            await self.disconnect()

        connection_state = None
        if isinstance(connector.current_state, states.Connected):
            connection_state = connector.current_state

        return connection_state

    async def disconnect(self):
        """
        Terminates a VPN connection.
        """
        connector = await self.get_vpn_connector()
        if not isinstance(connector.current_state, states.Disconnected):
            async with _wait_for_event(connector,
                                       event_types=[ConnectionStateEnum.DISCONNECTED]):
                await self._disconnect()

    async def login(self, username: str,
                    get_password: Callable[[], str],
                    get_2fa: Callable[[], str]):
        """
        Logs the user in.
        :param username:
        :param get_password: A callable that will return the account password
        :param get_2fa: A callable that will return the two factor
            authentication token if invoked.
        """
        if self._api.is_user_logged_in():
            raise SignoutRequiredError

        password = get_password()
        login_result = await self._api.login(username, password)
        if not login_result.authenticated:
            raise AuthenticationFailedError

        try:
            while login_result.twofa_required:
                login_result = await self._api.submit_2fa_code(get_2fa())
        except ProtonAPIAuthenticationNeeded as exc:
            raise Authentication2FAFailedError from exc

    async def logout(self):
        """
        Logs the user out.
        """
        if (await self.get_vpn_connector()).is_connection_active:
            await self.disconnect()

        await self._api.logout()

    async def get_updated_server_list(self) -> ServerList:
        """Returns an always-up-to-date server list."""
        return await self._api.refresher.get_up_to_date_server_list()

    async def is_serverlist_expired(self) -> bool:
        """Returns whether the caches serverlist has expired"""
        cached_server_list = self._api.server_list
        return (cached_server_list is None
                or cached_server_list.expired
                or cached_server_list.loads_expired)

    async def get_vpn_connector(self) -> VPNConnector:
        """Return the object that handles vpn connection and disconnection"""
        vpn_connector = await self._api.get_vpn_connector()
        return vpn_connector

    def validate_country_input(
            self,
            country: str,
    ) -> str:
        """Ensures that the passed country code ISO or country name is
        a valid value.

        Args:
            country (str): country ISO code or country name in plain english.

        Raises:
            CountryCodeError: if the passed country code does not exist
            CountryNameError: if the passed country name does not exist

        Returns:
            str: returns country code for a valid country
        """
        country_code = None

        if len(country) == 2:
            # assume the user specified a country code
            country_code = validate_country_code(country)
            if not country_code:
                raise CountryCodeError
        else:
            # assume user specified a country name
            country_code = get_country_code_for_name(country)
            if not country_code:
                raise CountryNameError

        return country_code

    async def _connect(self, server: LogicalServer):
        vpn_server = (await self.get_vpn_connector()).get_vpn_server(
            server, await self._api.refresher.get_up_to_date_client_config()
        )

        settings = await self._api.load_settings()

        await (await self.get_vpn_connector()).connect(
            vpn_server,
            protocol=settings.protocol
        )

    async def _disconnect(self):
        await (await self.get_vpn_connector()).disconnect()



"""
Server/Connection related commands.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from asyncio import CancelledError
from typing import Optional

import click

from proton.vpn.cli._cli_constants import PROGRAM_NAME
from proton.vpn.cli.core.exceptions import \
    AuthenticationRequiredError, \
    CountryCodeError, \
    CountryNameError, \
    RequiresHigherTierError, \
    VPNConnection2FARequiredError
from proton.vpn.cli.core.run_async import run_async
from proton.vpn.cli.core.controller import Controller, DEFAULT_CLI_NAME
from proton.vpn.cli.core.wait_for_current_tasks import wait_for_current_tasks
from proton.vpn.session.exceptions import ServerNotFoundError
from proton.vpn.session.servers.types import LogicalServer, ServerFeatureEnum
from proton.vpn.cli.commands.account import SIGNIN_COMMAND
from proton.vpn.cli.commands.command_utils import \
    inform_that_expired_serverlist_will_be_updated_if_necessary

CONNECT_COMMAND = "connect"
DISCONNECT_COMMAND = "disconnect"
STATUS_COMMAND = "status"
SERVERLIST_COMMAND = "servers"
SERVER_NAME_ARGUMENT = "SERVER_NAME"


class FailedConnection(click.ClickException):
    """When attempting to establish a connection, it fails
    """


def _print_usage_error(msg: str):
    raise click.UsageError(msg)


@click.command(
    name=CONNECT_COMMAND,
    epilog=f"""\b
Examples:
  # Quick connection
  {PROGRAM_NAME} {CONNECT_COMMAND}                     # Fastest server globally
\b
  # Location selection
  {PROGRAM_NAME} {CONNECT_COMMAND} --country US        # Fastest in United States
  {PROGRAM_NAME} {CONNECT_COMMAND} --city "New York"   # Fastest in New York
  {PROGRAM_NAME} {CONNECT_COMMAND} IT#23               # Specific server
\b
  # Feature-based selection
  {PROGRAM_NAME} {CONNECT_COMMAND} --p2p               # Fastest P2P server
  {PROGRAM_NAME} {CONNECT_COMMAND} --country IT --p2p  # Fastest P2P in Italy
\b
  # Extra secure connections
  {PROGRAM_NAME} {CONNECT_COMMAND} --securecore        # Secure Core routing
  {PROGRAM_NAME} {CONNECT_COMMAND} --tor               # Tor over VPN"""
)
@click.pass_context
@click.argument(SERVER_NAME_ARGUMENT, required=False)
@click.option(
    "--country",
    default=None,
    help="""\b
            Connect to fastest server in specified country
            Country code (US, GB, DE) or full name ("United States")""")
@click.option(
    "--city",
    default=None,
    help="""\b
            Connect to fastest server in specified city
            City name (use quotes for multi-word: \"New York\")""")
@click.option('--p2p', is_flag=True, help="Connect to fastest P2P-optimized server")
@click.option("-sc", "--securecore", is_flag=True, help="Connect to fastest Secure Core server")
@click.option("--tor", is_flag=True, help="Connect to fastest Tor server")
@click.option("--random", is_flag=True, help="Connect to random available server")
@run_async
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
# pylint: disable=too-many-branches
async def connect(
    ctx,
    server_name: Optional[str],
    city: Optional[str],
    country: Optional[str],
    p2p: bool,
    securecore: bool,
    tor: bool,
    random: bool
):
    """
    Connect to Proton VPN server.

    Arguments:

    SERVER_NAME Connect to specific server by ID (e.g., IT#23)
    """
    controller = await Controller.create(params=ctx.obj, click_ctx=ctx)
    # Silence cancelled exceptions raised by tasks we don't need to wait for after connection.
    # For example, some tasks are usually created to process a second Connected state broadcasted
    # to signal that the VPN server successfully applied the requested connection features.
    controller.set_uncaught_exceptions_to_absorb([CancelledError])
    server = None
    connection_state = None
    requested_features = _compose_requested_features(p2p, securecore, tor)

    await inform_that_expired_serverlist_will_be_updated_if_necessary(controller)

    # attempt to find a satisfactory server, and connect to it
    try:
        server = await controller.find_logical_server(
            server_name,
            country,
            city,
            requested_features,
            random
        )
        if not server:
            _print_usage_error("No servers found matching criteria. Try broadening your filters.")
            return

        connection_state = await controller.connect(server)

    except AuthenticationRequiredError:
        _print_usage_error(
            "Authentication required."
            f"Please sign in with '{controller.program_name} {SIGNIN_COMMAND}' before connecting."
        )
    except ServerNotFoundError as excp:
        if server_name:
            msg = f"Invalid server ID '{server_name}'. " \
                  "Please use a valid server ID from the server list."
        elif city:
            msg = f"City '{city}' not found or no servers available."
        else:
            msg = str(excp)

        _print_usage_error(msg)
    except CountryCodeError:
        _print_usage_error(f"Invalid country code '{country}'. Please use a valid country code.")
    except CountryNameError:
        _print_usage_error(f"Invalid country name '{country}'. Please use a valid country name.")
    except RequiresHigherTierError:
        free_user = controller.user_tier == 0
        if free_user:
            _display_free_user_limitation(
                controller,
                server_name,
                city,
                country,
                requested_features,
                random
            )
    except VPNConnection2FARequiredError as exc:
        raise click.ClickException(
            "2FA Required: You are connected to the VPN, but all traffic is blocked.\n"
            "You need to go to the authentication page provided by security and authenticate"
            " with your hardware key.\nAfter that, the traffic will be enabled."
        ) from exc

    if connection_state:
        # notify user of successful connection and server details
        current_connection = connection_state.context.connection
        connection_details = connection_state.context.event.context.connection_details
        server_ip = connection_details.server_ipv4 if connection_details else None
        click.echo(
            f"Connected to {current_connection.server_name} "
            f"in {_get_most_specific_server_location(server)}. "
        )
        if server_ip:
            click.echo(f"Your new IP address is {server_ip}.")

        await _display_relevant_server_capabilities(controller, server)

        protocol = (await controller.get_settings()).protocol
        _display_openvpn_warning_if_necessary(protocol)
    elif server:
        # we found a server but the connection failed
        raise FailedConnection(
            "Connection failed. "
            "Try connecting to a different server or check your network settings."
        )


@click.command(
    name=DISCONNECT_COMMAND,
    epilog=f"""\b
Examples:
  {PROGRAM_NAME} {DISCONNECT_COMMAND}
\b
Check connection status:
  {PROGRAM_NAME} {STATUS_COMMAND}"""
)
@click.pass_context
@run_async
async def disconnect(ctx):
    """Disconnect from current VPN server.

        \b
        This command will:
          - Terminate active VPN connection
          - Restore original network configuration"""
    controller = await Controller.create(params=ctx.obj, click_ctx=ctx)
    await controller.disconnect()
    click.echo("Disconnected.")

    # wait for post-disconnect notification killswitch implementation setting
    await wait_for_current_tasks()


@click.command(
    name=STATUS_COMMAND,
    epilog=f"""\b
            Examples:
              {PROGRAM_NAME} {STATUS_COMMAND}"""
)
@click.pass_context
@run_async
async def status(ctx):
    """\b
    Show current VPN connection status.
    \b
    Displays:
      - Connection status (Connected/Disconnected)
      - Server name and location (if connected)
      - Server load percentage (if connected)
      - Connection protocol (if connected)
    """
    controller = await Controller.create(params=ctx.obj, click_ctx=ctx)
    connection = (await controller.get_vpn_connector()).current_state.context.connection
    server_name = connection.server_name if connection else None
    status_lines = []
    if server_name:
        await inform_that_expired_serverlist_will_be_updated_if_necessary(controller)
        server_list = await controller.get_updated_server_list()
        server = server_list.get_by_name(server_name)
        server_location = _get_most_specific_server_location(server)
        status_lines.extend([
            "Status: Connected",
            f"Server: {server_name} in {server_location}",
            f"Load: {server.load}%",
            f"Protocol: {connection.protocol}"
        ])
    else:
        status_lines.extend([
            "Status: Disconnected"
        ])

    click.echo("\n".join(status_lines))


@click.command(
    name=SERVERLIST_COMMAND,
)
@run_async
async def servers():
    """
    View available servers. Prints the link to the full server list on protonvpn.com.
    """
    click.echo("To view detailed server information including specific server IDs, "
               "visit:  https://account.proton.me/vpn/WireGuard")


def _get_most_specific_server_location(server: LogicalServer) -> str:
    has_secure_core = ServerFeatureEnum.SECURE_CORE in server.features
    if has_secure_core and server.city:
        return f"{server.city}, via {server.entry_country_name}"

    if server.city:
        return f"{server.city}, {server.entry_country_name}"

    return server.entry_country_name


async def _display_relevant_server_capabilities(
    controller: Controller,
    server: LogicalServer
):
    settings = await controller.get_settings()
    if settings.features.port_forwarding:
        port_forwarding_enabled_server = ServerFeatureEnum.P2P in server.features
        if port_forwarding_enabled_server:
            click.echo(
                "\nPort forwarding is active on this server.\n"
                "To get your forwarded port, run the natpmpc setup script.\n"
                "Guide: https://protonvpn.com/support/port-forwarding-manual-setup#linux"
            )
        else:
            click.echo(
                "\nNote: Port forwarding is enabled but this server does not support it.\n"
                "Connect to a P2P server to use port forwarding:"
                f"{controller.program_name} {CONNECT_COMMAND} --p2p"
            )


OPENVPN_UDP = "openvpn-udp"
OPENVPN_TCP = "openvpn-tcp"


def _display_openvpn_warning_if_necessary(protocol: str):
    if protocol in [OPENVPN_UDP, OPENVPN_TCP]:
        click.echo(
            "\nOpenVPN is not fully supported in CLI and you may experience instability. "
            "For best results, use WireGuard."
        )


def _compose_requested_features(
    p2p: bool,
    securecore: bool,
    tor: bool
) -> ServerFeatureEnum:
    requested_features: ServerFeatureEnum = 0
    if p2p:
        requested_features |= ServerFeatureEnum.P2P
    if securecore:
        requested_features |= ServerFeatureEnum.SECURE_CORE
    if tor:
        requested_features |= ServerFeatureEnum.TOR

    return requested_features


# pylint: disable=too-many-arguments
def _display_free_user_limitation(
    controller: Controller,
    server_name: Optional[str],
    city: Optional[str],
    country: Optional[str],
    requested_features: Optional[ServerFeatureEnum],
    random: bool
):
    free_user = controller.user_tier == 0
    if not free_user:
        return

    # when specifying a server name, the user requires a paying tier
    if server_name:
        proton_cli_name = controller.program_name or DEFAULT_CLI_NAME
        _print_usage_error(
            f"Server selection by ID is not available on the free plan."
            f" Please use '{proton_cli_name} {CONNECT_COMMAND}' to connect "
            "to available free servers or upgrade to access all servers."
        )
        return

    # when specifying a country or city, the user requires a paying tier
    if country or city:
        proton_cli_name = controller.program_name or DEFAULT_CLI_NAME
        _print_usage_error(
            "Location selection is not available on the free plan. "
            f"Please use '{proton_cli_name} {CONNECT_COMMAND}' to connect "
            "to available free servers or upgrade to choose your location."
        )
        return

    # when specifying a server feature, the user requires a paying tier
    requested_feature_type = None
    if requested_features & ServerFeatureEnum.P2P != 0:
        requested_feature_type = "P2P"
    elif requested_features & ServerFeatureEnum.SECURE_CORE != 0:
        requested_feature_type = "Secure Core"
    elif requested_features & ServerFeatureEnum.TOR != 0:
        requested_feature_type = "Tor"

    if requested_feature_type:
        proton_cli_name = controller.program_name or DEFAULT_CLI_NAME
        _print_usage_error(
            f"{requested_feature_type} servers are not available on the free plan. "
            f"Please use '{proton_cli_name} {CONNECT_COMMAND}' to connect "
            "to available free servers "
            f"or upgrade to to access {requested_feature_type} servers."
        )
        return

    # when requesting a random server, the user requires a paying tier
    if random:
        proton_cli_name = controller.program_name or DEFAULT_CLI_NAME
        _print_usage_error(
            "Random selection is not available on the free plan. "
            f"Please use '{proton_cli_name} {CONNECT_COMMAND}' "
            "to connect to available free servers."
        )
        return



"""
Helper for async commands.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
import functools


def run_async(func):
    """
    Wraps an async function in a call to 'asyncio.run'.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))
    return wrapper



"""
Delegate for handling click errors.

Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import re
from typing import Optional

import click
from click.exceptions import \
    BadParameter, \
    ClickException, \
    MissingParameter, \
    UsageError

from proton.vpn.cli._cli_constants import HELP_OPTION
from proton.vpn.cli.commands.feature_setting_definitions import ALL_FEATURES


_BAD_CMD_RE = re.compile(r"No such command '([^']+)'")
_BAD_CHOICE_RE = re.compile(r": '([^']+)' is not ")


class ClickExceptionHandler:  # pylint: disable=too-few-public-methods
    """Handles exceptions generated by click when parsing and executing commands"""

    @staticmethod
    def handle_error(exception: Exception) -> bool:  # pylint: disable=too-many-return-statements
        """
        Handles the provided error and returns True.
        Returns False otherwise.
        """
        # MissingParameter and BadParameter are subclasses of UsageError,
        # so they must be checked before the generic UsageError handler.
        if isinstance(exception, MissingParameter):
            return ClickExceptionHandler._handle_missing_parameter(exception)

        if isinstance(exception, BadParameter):
            match = _BAD_CHOICE_RE.search(exception.format_message())
            bad_value = match.group(1) if match else "?"
            return ClickExceptionHandler._handle_bad_parameter(exception, bad_value)

        if isinstance(exception, UsageError):
            if exception.ctx and isinstance(exception.ctx.command, click.Group):
                match = _BAD_CMD_RE.search(exception.format_message())
                bad_cmd = match.group(1) if match else None
                is_root_cmd = exception.ctx.parent is None

                if is_root_cmd and not bad_cmd:
                    # Root command with no input given
                    # show the quickstart screen.
                    exception.show()
                    return True

                return ClickExceptionHandler._handle_no_such_command(exception, bad_cmd)

            lines = [
                f"Error: {exception.format_message()}\n",
            ]
            if exception.ctx:
                lines.append(
                    f"Try '{exception.ctx.command_path} {HELP_OPTION}' for more information."
                )
            click.echo("\n".join(lines), err=True)
            return True

        if isinstance(exception, ClickException):
            error_message = f"Error: {exception.format_message()}\n"
            click.echo(error_message, err=True)
            return True

        return False

    @staticmethod
    def _handle_no_such_command(exc: UsageError, bad_cmd: Optional[str]) -> bool:
        ctx = exc.ctx
        lines = []
        parsed_command_name = ctx.info_name
        if bad_cmd:
            lines.append(
                f"Error: Unknown action '{bad_cmd}' for '{parsed_command_name}' command.\n"
            )
        else:
            lines.append(f"Error: Missing action for '{parsed_command_name}' command.\n")

        subcommands = ClickExceptionHandler._list_subcommands(ctx)
        if subcommands:
            lines.append("Available actions:")
            col_width = max(len(name) for name, _ in subcommands) + 2
            for name, short_help in subcommands:
                lines.append(f"  {name:<{col_width}} {short_help}")

        lines.append(f"\nTry '{ctx.command_path} {HELP_OPTION}' for more information.")
        click.echo("\n".join(lines), err=True)
        return True

    @staticmethod
    def _handle_missing_parameter(exc: MissingParameter) -> bool:
        ctx = exc.ctx
        param = exc.param

        if param is None or ctx is None:
            return False

        if isinstance(param.type, click.Choice):
            parsed_command_name = ctx.info_name
            lines = [
                f"Error: Missing value for '{parsed_command_name}' setting.\n",
                *ClickExceptionHandler._format_valid_values(ctx.command.name, param.type.choices),
                f"\nTry '{ctx.command_path} {HELP_OPTION}' for examples.",
            ]
        else:
            param_label = param.human_readable_name or param.name or "argument"
            lines = [
                f"Error: Missing {param_label.lower()} argument.\n",
                f"Usage: {ctx.command_path} <{param_label.upper()}>\n",
                f"Try '{ctx.command_path} {HELP_OPTION}' for examples.",
            ]

        click.echo("\n".join(lines), err=True)
        return True

    @staticmethod
    def _handle_bad_parameter(exc: BadParameter, bad_value: str) -> bool:
        ctx = exc.ctx
        param = exc.param

        if param is None or ctx is None:
            return False

        if isinstance(param.type, click.Choice):
            parsed_command_name = ctx.info_name
            lines = [
                f"Error: Invalid value '{bad_value}' for '{parsed_command_name}'.\n",
                *ClickExceptionHandler._format_valid_values(ctx.command.name, param.type.choices),
                f"\nTry '{ctx.command_path} {HELP_OPTION}' for examples.",
            ]
        else:
            lines = [
                f"Error: {exc.format_message()}\n",
                f"Try '{ctx.command_path} {HELP_OPTION}' for more information.",
            ]

        click.echo("\n".join(lines), err=True)
        return True

    @staticmethod
    def _format_valid_values(command_name: str, choices: tuple[str, ...]) -> list[str]:
        """
        Returns lines for the 'Valid values' block.

        When the command matches a known feature setting, returns a multi-line
        block with value/description pairs sourced from the feature definition.
        Falls back to a single comma-separated line for unrecognised commands.
        """
        # maps a feature's command name to its list of value,help pairs
        feature_to_value_help = {
            feature.command: feature.value_to_help
            for feature in ALL_FEATURES
            if feature.value_to_help
        }
        value_help = feature_to_value_help.get(command_name)
        if value_help:
            return [
                "Valid values:",
                *[
                    f"  {value}  {description}"
                    for value, description in value_help.items()
                ],
            ]
        return [f"Valid values: {', '.join(choices)}"]

    @staticmethod
    def _list_subcommands(ctx: click.Context) -> list[tuple[str, str]]:
        """Returns (name, short_help) pairs for all visible subcommands."""
        result = []
        cmd = ctx.command
        if isinstance(cmd, click.Group):
            for name in cmd.list_commands(ctx):
                subcmd = cmd.get_command(ctx, name)
                if subcmd and not subcmd.hidden:
                    short_help = subcmd.get_short_help_str(limit=100)
                    result.append((name, short_help))
        return result




"""
Account/Authentication related commands.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import getpass

import click

from proton.vpn.cli._cli_constants import PROGRAM_NAME
from proton.vpn.cli.core.controller import Controller
from proton.vpn.cli.core.exceptions import \
    Authentication2FAFailedError, \
    AuthenticationFailedError, \
    SignoutRequiredError
from proton.vpn.cli.core.run_async import run_async


SIGNIN_COMMAND = "signin"
SIGNOUT_COMMAND = "signout"


@click.command(
    name=SIGNIN_COMMAND,
    epilog=f"""\b
           Examples:
             {PROGRAM_NAME} {SIGNIN_COMMAND} user@proton.me"""
)
@click.argument('username')
@click.pass_context
@run_async
async def signin(ctx, username: str):
    """
    Sign in to Proton VPN with your credentials.

    Arguments:

    USERNAME Proton account username
    """
    controller = await Controller.create(params=ctx.obj, click_ctx=ctx)
    try:
        await controller.login(
            username,
            getpass.getpass,
            lambda: getpass.getpass("2FA Token: ")
        )
        click.echo(f"Successfully signed in as '{controller.account_name}'")
    except SignoutRequiredError as exc:
        raise click.ClickException(
            "Already signed in, please sign out first before changing accounts."
        ) from exc
    except AuthenticationFailedError as exc:
        raise click.ClickException(
            "Authentication failed. Please check your username and password and try again."
        ) from exc
    except Authentication2FAFailedError as exc:
        raise click.ClickException(
            "2FA Authentication failed. Please try again."
        ) from exc


@click.command(
    name=SIGNOUT_COMMAND,
    epilog=f"""\b
           Examples:
             {PROGRAM_NAME} {SIGNOUT_COMMAND}
           \b
           To sign in again:
             {PROGRAM_NAME} {SIGNIN_COMMAND}"""
)
@click.pass_context
@run_async
async def signout(ctx):
    """Sign out from Proton VPN and clear local credentials."""
    controller = await Controller.create(params=ctx.obj, click_ctx=ctx)
    was_connected = await controller.is_connection_active()
    await controller.logout()
    completion_message = "You have been successfully signed out."
    if was_connected:
        completion_message = "VPN connection terminated and you've been successfully signed out."

    click.echo(completion_message)


@click.command()
@click.pass_context
@run_async
async def info(ctx):
    """Display your Proton VPN account information."""
    controller = await Controller.create(params=ctx.obj, click_ctx=ctx)
    click.echo(f"Account: '{controller.account_name}'")



"""
Set features commands.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import click
from tabulate import tabulate

from proton.vpn.cli._cli_constants import PROGRAM_NAME
from proton.vpn.cli.core.run_async import run_async
from proton.vpn.cli.core.controller import Controller
from proton.vpn.cli.core.exceptions import \
    AuthenticationRequiredError, \
    RequiresHigherTierError, \
    InvalidDNS
from proton.vpn.cli.commands.account import SIGNIN_COMMAND, SIGNOUT_COMMAND
from proton.vpn.cli.commands.feature_setting_definitions import \
    ALL_FEATURES, \
    BOOL_FEATURES, \
    ClickFeature, \
    ToggleType, \
    KillSwitchType, \
    NetshieldType, \
    CUSTOM_DNS_FEATURE, \
    IPV6_FEATURE, \
    KILLSWITCH_FEATURE, \
    NETSHIELD_FEATURE, \
    PORT_FORWARDING_FEATURE
from proton.vpn.cli.commands.server import DISCONNECT_COMMAND


CONFIG_COMMAND = "config"
SET_COMMAND = "set"
SETTINGS_LIST_COMMAND = "list"


def _raise_error_auth_required(controller: Controller, action: str) -> None:
    raise click.UsageError(
        f"Authentication required to {action} feature status. "
        f"Please sign in with '{controller.program_name} {SIGNIN_COMMAND}'"
    )


def _raise_error_requires_higher_tier(feature_human_friendly_name: str) -> None:
    raise click.UsageError(
        f"{feature_human_friendly_name} feature is not available "
        "on your current subscription plan. "
        "Please upgrade to access this feature."
    )


def _print_success_message(
    feature: ClickFeature,
    value: str,
    is_connection_active: bool = False
):
    msg = f"{feature.human_friendly_name} has been set to {value}"

    if feature.requires_restart and is_connection_active:
        msg += ", please establish a new VPN connection for " \
            "changes to take effect."

    msg += _feature_specific_success_postfix(feature, value)

    click.echo(msg)


def _feature_specific_success_postfix(
    feature: ClickFeature,
    value: str,
):
    if feature.command == PORT_FORWARDING_FEATURE.command and \
       value == ToggleType.get_human_friendly_state_string(True):
        return \
            "\nWhen connected to a P2P server, the VPN will request a forwarded port.\n" \
            "To receive and maintain your port, follow the setup guide:\n" \
            "https://protonvpn.com/support/port-forwarding-manual-setup#linux"

    return ""


def _build_config_epilog() -> str:
    lines = [f"""\b
Examples:
  {PROGRAM_NAME} {CONFIG_COMMAND} {SETTINGS_LIST_COMMAND}
  {PROGRAM_NAME} {CONFIG_COMMAND} {SET_COMMAND} {NETSHIELD_FEATURE.command} malware-ads-trackers
  {PROGRAM_NAME} {CONFIG_COMMAND} {SET_COMMAND} {KILLSWITCH_FEATURE.command} standard
  {PROGRAM_NAME} {CONFIG_COMMAND} {SET_COMMAND} {IPV6_FEATURE.command} on
\b
Available Settings:"""]

    for feature in ALL_FEATURES:
        lines.append(f"  {feature.command:<22} {feature.available_setting_description}")

    lines.extend([f"""\b
For setting-specific help:
  {PROGRAM_NAME} {CONFIG_COMMAND} {SET_COMMAND} <setting> --help"""])

    return "\n".join(lines)


@click.group(
    name=CONFIG_COMMAND,
    no_args_is_help=False,
    epilog=_build_config_epilog()
)
def config():
    """Configure Proton VPN settings."""


@config.group(
    name=SET_COMMAND,
    no_args_is_help=False
)
def set_group():
    """Change a specific setting."""


def _format_setting_epilog(feature: ClickFeature) -> str:
    values_with_help = []
    for value_str in feature.click_type.to_list_of_str():
        values_with_help.append(f"  {value_str}   {feature.value_to_help[value_str]}")

    return feature.help_epilog.format(
        feature_values="\n".join(values_with_help),
        program_name=PROGRAM_NAME,
        config_command=CONFIG_COMMAND,
        set_command=SET_COMMAND,
        settings_list_command=SETTINGS_LIST_COMMAND,
        feature_command=feature.command,
    )


def _register_bool_feature_command(group: click.Group, feature: ClickFeature):
    @group.command(
        name=feature.command,
        help=feature.help_description,
        epilog=_format_setting_epilog(feature)
    )
    @click.argument("state", type=click.Choice(ToggleType.to_list_of_str(), case_sensitive=False))
    @click.pass_context
    @run_async
    async def _bool_command(ctx: click.Context, state: str) -> None:
        controller = await Controller.create(params=ctx.obj, click_ctx=ctx)
        toggle_value = ToggleType.from_str(state)
        try:
            await controller.save_feature_setting(feature, toggle_value)
        except AuthenticationRequiredError:
            _raise_error_auth_required(controller, action="set")
        except RequiresHigherTierError:
            _raise_error_requires_higher_tier(feature.human_friendly_name)
        else:
            _print_success_message(
                feature,
                ToggleType.get_human_friendly_state_string(toggle_value),
                is_connection_active=await controller.is_connection_active()
            )


for _feature in BOOL_FEATURES:
    _register_bool_feature_command(set_group, _feature)


@set_group.command(
    name=KILLSWITCH_FEATURE.command,
    help=KILLSWITCH_FEATURE.help_description,
    epilog=_format_setting_epilog(KILLSWITCH_FEATURE)
)
@click.argument("mode", type=click.Choice(KillSwitchType.to_list_of_str(), case_sensitive=False))
@click.pass_context
@run_async
async def killswitch_command(ctx: click.Context, mode: str) -> None:
    """Configure Kill Switch to block internet if VPN connection drops."""
    controller = await Controller.create(params=ctx.obj, click_ctx=ctx)
    if await controller.is_connection_active():
        raise click.UsageError(
            "Disconnect before changing Kill Switch. "
            f"Run '{controller.program_name} {DISCONNECT_COMMAND}' first."
        )

    killswitch_value = KillSwitchType.from_str(mode)

    try:
        await controller.save_feature_setting(KILLSWITCH_FEATURE, killswitch_value)
    except AuthenticationRequiredError:
        _raise_error_auth_required(controller, action="set")
    else:
        _print_success_message(
            KILLSWITCH_FEATURE,
            KillSwitchType.get_human_friendly_state_string(killswitch_value)
        )


@set_group.command(
    name=NETSHIELD_FEATURE.command,
    help=NETSHIELD_FEATURE.help_description,
    epilog=_format_setting_epilog(NETSHIELD_FEATURE)
)
@click.argument(
    "mode",
    type=click.Choice(NetshieldType.to_list_of_str(), case_sensitive=False),
)
@click.pass_context
@run_async
async def netshield_command(ctx: click.Context, mode: str) -> None:
    """Configure NetShield ad-blocking and malware protection."""
    controller = await Controller.create(params=ctx.obj, click_ctx=ctx)
    netshield_value = NetshieldType.from_str(mode)

    try:
        await controller.save_feature_setting(NETSHIELD_FEATURE, netshield_value)
    except AuthenticationRequiredError:
        _raise_error_auth_required(controller, action="set")
    except RequiresHigherTierError:
        _raise_error_requires_higher_tier(NETSHIELD_FEATURE.human_friendly_name)
    else:
        _print_success_message(
            NETSHIELD_FEATURE,
            f"'{NetshieldType.get_human_friendly_state_string(netshield_value)}'"
        )


@set_group.command(
    name=CUSTOM_DNS_FEATURE.command,
    help=CUSTOM_DNS_FEATURE.help_description,
    epilog=_format_setting_epilog(CUSTOM_DNS_FEATURE)
)
@click.argument("state", type=click.Choice(ToggleType.to_list_of_str(), case_sensitive=False))
@click.option("--dns", "dns_csv", help="Comma-separated DNS server IPs (IPv4 or IPv6)")
@click.pass_context
@run_async
async def custom_dns_command(ctx: click.Context, state: str, dns_csv: str | None) -> None:
    """Configure custom DNS servers."""
    controller = await Controller.create(params=ctx.obj, click_ctx=ctx)
    toggle_value = ToggleType.from_str(state)

    parsed_dns_ips = []
    dns_list = [x.strip() for x in dns_csv.split(",") if x.strip()] if dns_csv else []

    if toggle_value is True:
        if not dns_list:
            raise click.UsageError(
                f"When enabling {CUSTOM_DNS_FEATURE.human_friendly_name} feature "
                "you must provide a list of comma separated DNS's."
            )

        try:
            parsed_dns_ips = controller.parse_dns_ips(dns_list)
        except InvalidDNS as excp:
            raise click.UsageError(
                f"Invalid DNS address '{excp.dns}'. Please provide a valid IPv4 address."
            )

    custom_dns = controller.to_custom_dns(toggle_value, parsed_dns_ips)

    try:
        await controller.save_feature_setting(CUSTOM_DNS_FEATURE, custom_dns)
    except AuthenticationRequiredError:
        _raise_error_auth_required(controller, action="set")
    except RequiresHigherTierError:
        _raise_error_requires_higher_tier(CUSTOM_DNS_FEATURE.human_friendly_name)
    else:
        _print_success_message(
            CUSTOM_DNS_FEATURE,
            ToggleType.get_human_friendly_state_string(toggle_value),
            is_connection_active=await controller.is_connection_active()
        )


@config.command(name=SETTINGS_LIST_COMMAND)
@click.pass_context
@run_async
async def list_features(ctx: click.Context):
    """Show current configuration for all settings."""
    controller = await Controller.create(params=ctx.obj, click_ctx=ctx)

    settings_table = []
    for feature in ALL_FEATURES:
        try:
            setting = await controller.get_feature_setting(feature)
        except AuthenticationRequiredError:
            _raise_error_auth_required(controller, action="view")

        feature_requires_upgrade =\
            not feature.available_on_free_tier and controller.user_on_free_tier

        setting_str = \
            "Upgrade to enable" if feature_requires_upgrade\
            else feature.click_type.to_str(setting)
        settings_table.append((feature.command, setting_str))

    table = tabulate(
        settings_table,
        headers=["Setting", "Value"],
        tablefmt="simple",
        stralign="left",
        numalign="right",
    )
    click.echo(f"\nCurrent configuration\n{table}\n")

    # determine footer guidance
    if controller.user_on_free_tier:
        program_name = controller.program_name
        click.echo(
            "To upgrade to VPN Plus visit: https://account.protonvpn.com/pricing\n"
            "After upgrading online:\n"
            "    Sign out and sign in again to activate your new plan:\n"
            f"    {program_name} {SIGNOUT_COMMAND} && {program_name} {SIGNIN_COMMAND}"
        )
    else:
        full_set_command_str = f"{controller.program_name} {CONFIG_COMMAND} {SET_COMMAND}"
        help_option_str = ctx.help_option_names[0]
        click.echo(
            f"Use '{full_set_command_str} <setting> <value>' to change settings.\n"
            f"Use '{full_set_command_str} <setting> {help_option_str}' for available values."
        )
